#include <string.h>
#include "zlib.h"
#include "lua.h"
#include "lauxlib.h"

static int cksum_common(lua_State* L, int defval,
    uLong (*cksum)(uLong, const Bytef*, uInt)) {
  size_t srclen;
  const Bytef* src=(const Bytef*)luaL_checklstring(L,1,&srclen);
  uLong val = luaL_optunsigned(L, 2, defval);
  val = cksum(val, src, srclen);
  lua_pushunsigned(L, val);
  return 1;
}

static int zlib_adler(lua_State* L) {
  return cksum_common(L, 1, adler32);
}

static int zlib_crc(lua_State* L) {
  return cksum_common(L, 0, crc32);
}

static int compres_common(lua_State* L, int fdeflate) {
  size_t srclen;
  const Bytef* src=(const Bytef*)luaL_checklstring(L,1,&srclen);
  int level = luaL_optint(L, 2, Z_DEFAULT_COMPRESSION);
  uLong dstlen = srclen;
  Byte* dst;
  int res;
  while (1)
  {
    dstlen *= 2;
    lua_settop(L, 1);
    dst = (Bytef*)lua_newuserdata(L, dstlen);
    if (fdeflate)
      res = compress2(dst, &dstlen, src, srclen, level);
    else
      res = uncompress(dst, &dstlen, src, srclen);
    switch (res)
    {
    case Z_OK:
      lua_pushlstring(L, (const char*)dst, dstlen);
      return 1;
    case Z_BUF_ERROR:
      continue;
    case Z_MEM_ERROR:
      luaL_error(L, "Erreur mmoire durant la %scompression", 
        fdeflate ? "" : "d");
    case Z_DATA_ERROR:
      luaL_error(L, "Donnes compresses corrompues");
    case Z_STREAM_ERROR:
      luaL_error(L, "Niveau de compression incorrect");
    default:
      luaL_error(L, "Erreur zlib n %d", res);
    }
  }
}

static int zlib_compress(lua_State* L) {
  return compres_common(L, 1);
}

static int zlib_uncompress(lua_State* L) {
  return compres_common(L, 0);
}

static int zlib_newsteam(lua_State* L) {
  z_streamp strm=(z_streamp)lua_newuserdata(L, sizeof(z_stream));
  memset(strm, 0, sizeof(z_stream));
  return 1;
}

static int deflate_common(lua_State* L, Byte* src, 
    size_t srclen, int fdeflate, int flush, int end_result) {
  z_streamp strm = (z_streamp)lua_touserdata(L, 1);
  int res = -1;
  luaL_Buffer buf;

  if(strm == NULL)
    luaL_error(L, "L'argument 1 doit tre un flux");
  strm->next_in = src;
  strm->avail_in = srclen;
  luaL_buffinit(L, &buf);
  while(strm->avail_in || res < end_result)
  {
    enum { SIZE_BUF = 1024 };
    strm->next_out = (Byte*)luaL_prepbuffsize(&buf, SIZE_BUF);
    strm->avail_out = SIZE_BUF;
    if(fdeflate)
      res = deflate(strm, flush);
    else
      res = inflate(strm, flush);
    if(res < 0)
      luaL_error(L, "Erreur %d  la %scompression", res, 
        fdeflate ? "" : "d");
    luaL_addsize(&buf, SIZE_BUF - strm->avail_out);
  }
  luaL_pushresult(&buf);
  return 1;
}

static int zlib_deflateInit(lua_State* L) {
  z_streamp strm = (z_streamp)lua_touserdata(L, 1);
  int level = luaL_optint(L, 2, Z_DEFAULT_COMPRESSION);
  deflateInit(strm, level);
  return 0;
}

static int zlib_deflate(lua_State* L) {
  size_t srclen;
  Bytef* src = (Bytef*)luaL_checklstring(L, 2, &srclen);
  return deflate_common(L, src, srclen, 1, Z_NO_FLUSH, Z_OK);
}

static const char* const flush_names[] = {
  "no", "partial", "sync", "full", "finish", "block", "trees"
};

static int zlib_deflateFlush(lua_State* L) {
  int flush = luaL_checkoption(L, 2, "finish", flush_names);
  return deflate_common(L, NULL, 0, 1, flush, 
    flush == Z_FINISH ? Z_STREAM_END : Z_OK);
}

static int zlib_deflateEnd(lua_State* L) {
  z_streamp strm = (z_streamp)lua_touserdata(L, 1);
  deflateEnd(strm);
  return 0;
}

static int zlib_inflateInit(lua_State* L) {
  z_streamp strm = (z_streamp)lua_touserdata(L, 1);
  inflateInit(strm);
  return 0;
}

static int zlib_inflate(lua_State* L) {
  size_t srclen;
  Bytef* src = (Bytef*)luaL_checklstring(L, 2, &srclen);
  return deflate_common(L, src, srclen, 0, Z_NO_FLUSH, Z_OK);
}

static int zlib_inflateFlush(lua_State* L) {
  int flush = luaL_checkoption(L, 2, "finish", flush_names);
  return deflate_common(L, NULL, 0, 0, flush, 
    flush == Z_FINISH ? Z_STREAM_END : Z_OK);
}

static int zlib_inflateEnd(lua_State* L) {
  z_streamp strm = (z_streamp)lua_touserdata(L, 1);
  inflateEnd(strm);
  return 0;
}

static const luaL_Reg zlib_fcts[] = {
  { "crc",          zlib_crc },
  { "adler",        zlib_adler },
  { "compress",     zlib_compress },
  { "uncompress",   zlib_uncompress },
  { "newstream",    zlib_newsteam },
  { "deflateInit",  zlib_deflateInit },
  { "deflate",      zlib_deflate },
  { "deflateFlush", zlib_deflateFlush },
  { "deflateEnd",   zlib_deflateEnd },
  { "inflateInit",  zlib_inflateInit },
  { "inflate",      zlib_inflate },
  { "inflateFlush", zlib_inflateFlush },
  { "inflateEnd",   zlib_inflateEnd },
  { NULL, NULL }
};

int luaopen_zlib(lua_State* L) {
  luaL_newlib(L, zlib_fcts);
  lua_pushstring(L, ZLIB_VERSION);
  lua_setfield(L, -2, "_VERSION");
  return 1;
}
