#include <string.h>
#include "zlib.h"
#include "lua.h"
#include "lauxlib.h"

static int cksum_common(lua_State* L, int defval,
    uLong (*cksum)(uLong, const Bytef*, uInt)) {
  size_t srclen;
  const Bytef* src=(const Bytef*)luaL_checklstring(L,1,&srclen);
  uLong val = luaL_optunsigned(L, 2, defval);
  val = cksum(val, src, srclen);
  lua_pushunsigned(L, val);
  return 1;
}

static int zlib_adler(lua_State* L) {
  return cksum_common(L, 1, adler32);
}

static int zlib_crc(lua_State* L) {
  return cksum_common(L, 0, crc32);
}

static int compres_common(lua_State* L, int fdeflate) {
  size_t srclen;
  const Bytef* src=(const Bytef*)luaL_checklstring(L,1,&srclen);
  int level = luaL_optint(L, 2, Z_DEFAULT_COMPRESSION);
  uLong dstlen = srclen;
  Byte* dst;
  int res;
  while (1)
  {
    dstlen *= 2;
    lua_settop(L, 1);
    dst = (Bytef*)lua_newuserdata(L, dstlen);
    if (fdeflate)
      res = compress2(dst, &dstlen, src, srclen, level);
    else
      res = uncompress(dst, &dstlen, src, srclen);
    switch (res)
    {
    case Z_OK:
      lua_pushlstring(L, (const char*)dst, dstlen);
      return 1;
    case Z_BUF_ERROR:
      continue;
    case Z_MEM_ERROR:
      luaL_error(L, "Erreur mmoire durant la %scompression", 
        fdeflate ? "" : "d");
    case Z_DATA_ERROR:
      luaL_error(L, "Donnes compresses corrompues");
    case Z_STREAM_ERROR:
      luaL_error(L, "Niveau de compression incorrect");
    default:
      luaL_error(L, "Erreur zlib n %d", res);
    }
  }
}

static int zlib_compress(lua_State* L) {
  return compres_common(L, 1);
}

static int zlib_uncompress(lua_State* L) {
  return compres_common(L, 0);
}

#define DEFLATE_META "ZLIB_Deflate_meta"
#define INFLATE_META "ZLIB_Inflate_meta"

static z_streamp zlib_newsteam(lua_State* L, 
  const char* metaname) {
  z_streamp strm=(z_streamp)lua_newuserdata(L, sizeof(z_stream));
  memset(strm, 0, sizeof(z_stream));
  luaL_setmetatable(L, metaname);
  return strm;
}

static int deflate_common(lua_State* L, Byte* src, 
    size_t srclen, int fdeflate, int flush, int end_result) {
  z_streamp strm = (z_streamp)luaL_checkudata(L, 1, 
    fdeflate ? DEFLATE_META : INFLATE_META);
  int res = -1;
  luaL_Buffer buf;

  strm->next_in = src;
  strm->avail_in = srclen;
  luaL_buffinit(L, &buf);
  while(strm->avail_in || res < end_result)
  {
    enum { SIZE_BUF = 1024 };
    strm->next_out = (Byte*)luaL_prepbuffsize(&buf, SIZE_BUF);
    strm->avail_out = SIZE_BUF;
    if(fdeflate)
      res = deflate(strm, flush);
    else
      res = inflate(strm, flush);
    if(res < 0)
      luaL_error(L, "Erreur %d  la %scompression", res, 
        fdeflate ? "" : "d");
    luaL_addsize(&buf, SIZE_BUF - strm->avail_out);
  }
  luaL_pushresult(&buf);
  return 1;
}

static int zlib_deflateInit(lua_State* L) {
  int level = luaL_optint(L, 1, Z_DEFAULT_COMPRESSION);
  z_streamp strm = zlib_newsteam(L, DEFLATE_META);
  deflateInit(strm, level);
  return 1;
}

static int zlib_deflate(lua_State* L) {
  size_t srclen;
  Bytef* src = (Bytef*)luaL_checklstring(L, 2, &srclen);
  return deflate_common(L, src, srclen, 1, Z_NO_FLUSH, Z_OK);
}

static const char* const flush_names[] = {
  "no", "partial", "sync", "full", "finish", "block", "trees"
};

static int zlib_deflateFlush(lua_State* L) {
  int flush = luaL_checkoption(L, 2, "finish", flush_names);
  return deflate_common(L, NULL, 0, 1, flush, 
    flush == Z_FINISH ? Z_STREAM_END : Z_OK);
}

static int zlib_deflateEnd(lua_State* L) {
  z_streamp strm=(z_streamp)luaL_checkudata(L, 1, DEFLATE_META);
  deflateEnd(strm);
  return 0;
}

static int zlib_deflateToString(lua_State* L) {
  lua_pushfstring(L, "ZLib deflater (%p)", lua_topointer(L, 1));
  return 1;
}


static int zlib_inflateInit(lua_State* L) {
  z_streamp strm = zlib_newsteam(L, INFLATE_META);
  inflateInit(strm);
  return 1;
}

static int zlib_inflate(lua_State* L) {
  size_t srclen;
  Bytef* src = (Bytef*)luaL_checklstring(L, 2, &srclen);
  return deflate_common(L, src, srclen, 0, Z_NO_FLUSH, Z_OK);
}

static int zlib_inflateFlush(lua_State* L) {
  int flush = luaL_checkoption(L, 2, "finish", flush_names);
  return deflate_common(L, NULL, 0, 0, flush, 
    flush == Z_FINISH ? Z_STREAM_END : Z_OK);
}

static int zlib_inflateEnd(lua_State* L) {
  z_streamp strm=(z_streamp)luaL_checkudata(L, 1, INFLATE_META);
  inflateEnd(strm);
  return 0;
}

static int zlib_inflateToString(lua_State* L) {
  lua_pushfstring(L, "ZLib inflater (%p)", lua_topointer(L, 1));
  return 1;
}
static const luaL_Reg zlib_fcts[] = {
  { "crc",        zlib_crc },
  { "adler",      zlib_adler },
  { "compress",   zlib_compress },
  { "uncompress", zlib_uncompress },
  { "deflater",   zlib_deflateInit },
  { "inflater",   zlib_inflateInit },
  { NULL, NULL }
};

static const luaL_Reg deflate_fcts[] = {
  { "add",        zlib_deflate },
  { "flush",      zlib_deflateFlush },
  { "__gc",       zlib_deflateEnd },
  { "__tostring", zlib_deflateToString },
  { NULL, NULL }
};

static const luaL_Reg inflate_fcts[] = {
  { "add",        zlib_inflate },
  { "flush",      zlib_inflateFlush },
  { "__gc",       zlib_inflateEnd },
  { "__tostring", zlib_inflateToString },
  { NULL, NULL }
};

static void zlib_createmeta (lua_State *L, 
  const char* metaname, const luaL_Reg* fcts) {
  luaL_newmetatable(L, metaname);
  lua_pushvalue(L, -1);
  lua_setfield(L, -2, "__index");
  luaL_setfuncs(L, fcts, 0);
  lua_pop(L, 1);
}

int luaopen_zlib(lua_State* L) {
  luaL_newlib(L, zlib_fcts);
  lua_pushstring(L, ZLIB_VERSION);
  lua_setfield(L, -2, "_VERSION");
  zlib_createmeta(L, DEFLATE_META, deflate_fcts);
  zlib_createmeta(L, INFLATE_META, inflate_fcts);
  return 1;
}
